@@ -36,9 +36,17 @@ def main(args):
     print("classes:", d.classes_)
     print("features:", d.features_)
     print("n. samples:", len(d.samples_))
-    print("  training:", len(d.split_.training_))
-    print("  validation:", len(d.split_.validation_))
-    print("  test:", len(d.split_.test_))
+    print("splits:")
+    for s in d.split_:
+        print(f'  "{s.split_name_}":')
+        print(f'    type: {s.split_type_}')
+        print(f'    n. samples: {len(s.samples_indices_)}')
+        print(f'    drop last: {s.drop_last_}')
+        print(f'    n. batches: {s.num_batches_}')
+        print(f'    last batch: {s.last_batch_}')
+        print(f'    no label: {s.no_label_}')
+    print(f"current split: {d.current_split_}")
+    print(f"task: {d.task_}")
     print("loading first sample image")
     print("  location:", d.samples_[0].location_)
     print("  label_path:", d.samples_[0].label_path_)
